with Unchecked_Deallocation;

package body Contenedores.Rojo_Negros is
	--------------------------
	-- Operaciones Internas --
	--------------------------	
	procedure Eliminar is new Unchecked_Deallocation(Nodo_Rojo_Negro, rbol_Rojo_Negro);

	procedure Rotar_simple_Rojo_Negro(
			A   : in out Rbol_Rojo_Negro;
			Rama: in integer
			) is
		AuxColor: Boolean;
		OtraRama: Integer := 3 - Rama;
		Hijo    : Rbol_Rojo_Negro := A.Hijos(Rama);
	begin
		-- Se intercambian los nodos
		A.Hijos(Rama)        := Hijo.Hijos(OtraRama);
		Hijo.Hijos(OtraRama) := A;
		A                    := Hijo;

		-- Se ajustan los colores
		Hijo         := A.Hijos(OtraRama);
		AuxColor     := A.Es_Rojo;
		A.Es_Rojo    := Hijo.es_Rojo;
		Hijo.Es_Rojo := AuxColor;
	end Rotar_Simple_Rojo_Negro;

	Procedure Rotar_Doble_Rojo_Negro(
			A   : in out Rbol_Rojo_Negro;
			Rama: in Integer) is
		OtraRama: Integer := 3 - Rama;
		Hijo    : Rbol_Rojo_Negro := A.Hijos(Rama);
		Nieto   : Rbol_Rojo_Negro := Hijo.Hijos(OtraRama);
		OtroHijo: Rbol_Rojo_Negro;
	begin
		-- Se intercambian los nodos
		A.Hijos(Rama)         := Nieto.Hijos(OtraRama);
		Hijo.Hijos(OtraRama)  := Nieto.Hijos(Rama);
		Nieto.Hijos(OtraRama) := A;
 	    Nieto.Hijos(Rama)     := Hijo;
		A                     := Nieto;

		-- Se ajustan los colores
		OtroHijo         := A.Hijos(OtraRama);
		A.Es_Rojo        := False;
		OtroHijo.Es_Rojo:= True;
	end Rotar_Doble_Rojo_Negro;

	-- Pone el color de un nodo a negro y el de sus hijos a rojo
	procedure Recolorar(A: in out Rbol_Rojo_negro) is
	begin
		A.Es_Rojo          := True;
		A.Hijos(1).ES_Rojo := False;
		A.Hijos(2).ES_Rojo := False;
	end Recolorar;

	-- Indica si los dos hijos de A son negros
	function Hijos_Negros(A: in Rbol_Rojo_negro) return Boolean is
	begin
		return (A.Hijos = (null, null)) or else 
			((A.Hijos(1) /= null and then not A.Hijos(1).Es_Rojo)
			and (A.Hijos(2) /= null and then not A.Hijos(2).Es_Rojo));
	end Hijos_Negros;

	procedure Insertar(
			Raz: in rbol_Rojo_Negro;
			A: in out rbol_Rojo_Negro;
			E: in TElemento;
			Propiedad_Rojo: in out boolean
			) is
		Rama: Integer;
		Padre, Tio, Hijo: rbol_Rojo_Negro;
	begin
		if A = null then
			-- El nuevo valor se inserta como hoja
			A := new Nodo_Rojo_Negro'(True,E, (null, null));
			Propiedad_Rojo := True;
		else
			-- Si E = A.Clave, no se inserta
			if E /= A.Clave then
				-- La insercin continua en el subrbol adecuado
				if E < A.Clave then
					Rama := 1;
				else
					Rama := 2;
				end if;

				Insertar(Raz, A.Hijos(Rama), E, Propiedad_Rojo);

				if Propiedad_Rojo then					
					if A.Es_Rojo then
						-- Hay que ascender otro nivel, a menos que
						-- se est en la raz
						if A = Raz then
							A.Es_Rojo      := False;
							Propiedad_Rojo := False;
						end if;
					else
						-- A es negro
						Padre := A.Hijos(Rama);
						-- Se ha resuelto el conflicto rojo-rojo?
						if Hijos_Negros(Padre) then
							Propiedad_Rojo := False;
						else
							Tio := A.Hijos(3 - Rama);
							-- El color del tio determina si se elige una
							-- recoloracin o una rotacin
							if Tio /= null and then Tio.Es_Rojo then
								Recolorar(A);
							else
								Hijo := Padre.Hijos(Rama);
								if (Hijo /= null and then Hijo.Es_Rojo)
									and Padre.ES_Rojo
								then
									Rotar_Simple_Rojo_Negro(A, Rama);
								else
									Rotar_Doble_Rojo_Negro(A, Rama);
								end if;
								-- Se han restablecido las propiedades del rbol
								Propiedad_Rojo := False;
							end if;
						end if;
					end if;
				end if;
			end if;
		end if;
	end Insertar;

	procedure Extraer(
			Raz: in rbol_Rojo_Negro;
			A   : in out rbol_Rojo_Negro;
			E   : in TElemento;
			Propiedad_Camino: in out Boolean
			) is

		Procedure Restaurar_propiedad_Camino(
	   			A: in out Rbol_Rojo_Negro;
				rama: in Integer
				) is
			Hermano, Alejado, Prximo: rbol_Rojo_Negro;
		begin		
			Hermano := A.Hijos(3 - Rama);
			if Hermano.Es_Rojo then
				-- Caso 1 del texto
				Rotar_Simple_Rojo_Negro(A, 3 - Rama);
				Hermano := A.Hijos(Rama).Hijos(3 - rama);
				if not Hermano.es_Rojo and Hijos_Negros(Hermano) then
					Hermano.Es_Rojo := True;
					A.Hijos(Rama).Es_Rojo := False;
				else
					Restaurar_propiedad_Camino(A.Hijos(Rama), Rama);
				end if;
			else
				Alejado := Hermano.Hijos(3 - Rama);
				Prximo := Hermano.Hijos(Rama);
				if Alejado /= null and then Alejado.Es_Rojo then
					-- Caso 3 del texto
					Rotar_simple_rojo_Negro(A, 3 - Rama);
					Alejado.Es_Rojo := False;
				else
					if (Alejado = null or else not Alejado.Es_Rojo) and
						Prximo.Es_Rojo
					then
						-- Caso 4 del texto
						Rotar_Simple_Rojo_Negro(A.Hijos(3 - Rama), Rama);
						Restaurar_propiedad_Camino(A, Rama);
					end if;
				end if;
			end if;
		end Restaurar_propiedad_Camino;

		procedure Extraer_Sucesor(
				Nodo_Extraer: in out rbol_Rojo_Negro;
				Nodo_Actual : in out rbol_Rojo_Negro;
				Propiedad_Camino: in out Boolean
				) is
			Hijo, Hermano: rbol_Rojo_Negro;
		begin
			if Nodo_Actual.Hijos(1) = null then
				-- Nodo_Actual es el sucesor
				Hijo               := Nodo_Actual.Hijos(2);
				Nodo_Extraer.Clave := Nodo_Actual.Clave;
				Propiedad_Camino   := True;

				if Nodo_Actual.Es_Rojo then
					Propiedad_Camino := False;
				else
					if Hijo /= null and then Hijo.Es_Rojo then
						Hijo.Es_Rojo := False;
						Propiedad_Camino := False;
					end if;
				end if;

				Eliminar(Nodo_Actual);
				Nodo_Actual := Hijo;
			else
				-- El sucesor es descendiente de Nodo_Actual
				Extraer_Sucesor(Nodo_Extraer, Nodo_Actual.Hijos(1), Propiedad_Camino);

				if Propiedad_Camino then
					Hermano := Nodo_Actual.Hijos(2);
					if not(hermano.Es_Rojo) and Hijos_Negros(Hermano) then
						-- Caso 2 del texto (Se produce tambin en el cuerpo
						-- de Extraer)
						Hermano.Es_Rojo := True;
						if Nodo_Actual.Es_Rojo then
							Nodo_Actual.Es_rojo := False;
							Propiedad_Camino   := False;
					    end if;
					else
						Restaurar_propiedad_Camino(Nodo_Actual, 1);
						Propiedad_Camino := False;
					end if;
				end if;
			end if;
		end Extraer_Sucesor;
		
		Rama: Integer;
		Hijo, hermano: rbol_Rojo_Negro; 
	begin
		if A /= null then
			-- Si E /= A.Clave, la extraccin se producir en
			-- el subrbol que corresponda
 	    	if E < A.Clave then
				Extraer(Raz, A.Hijos(1), E, Propiedad_Camino);
				Rama := 1;
			else
				if A.Clave < E then
					Extraer(Raz, A.Hijos(2), E, Propiedad_Camino);
					Rama := 2;
				else
					-- Se ha encontrado el nodo a extraer
					-- Tiene grado 0 0 1?
					if A.Hijos(1) = null or else A.Hijos(2) = null then
						if A.Hijos(1) = null then
							Rama := 2;
						else
							Rama := 1;
						end if;
						Hijo := A.Hijos(Rama);
						Propiedad_Camino := True;

						if A.Es_Rojo or A = Raz then
							Propiedad_Camino := False;
						else
							if Hijo /= null and then Hijo.Es_Rojo then
								Hijo.Es_Rojo := False;
								Propiedad_Camino := False;
							end if;
						end if;
						
						Eliminar(A);
						A := Hijo;
						return;
					else
						-- Tiene grado 2, hay que extraer su sucesor
						-- en simtrico
						Extraer_Sucesor(A, A.Hijos(2), Propiedad_Camino);
						Rama := 2;
					end if;
				end if;
			end if;

			-- Un cambio de altura require ajustar el factor de equilibrio
			if Propiedad_Camino then
				Hermano := A.Hijos(3 - Rama);
				if not(Hermano.Es_Rojo) and Hijos_Negros(Hermano) then
					-- Caso 2 del texto (Se produce tambin dentro de
					-- Extraer_Sucesor)
					Hermano.Es_Rojo := True;
					if A.Es_Rojo then
						A.Es_Rojo := False;
						Propiedad_Camino := False;
					end if;
				else
					Restaurar_Propiedad_Camino(A, Rama);
					Propiedad_Camino := False;
				end if;
			end if;
		end if;
	end Extraer;

	function Buscar(A: rbol_Rojo_Negro; E: in TElemento) 
			return  rbol_Rojo_Negro is
	begin
		if A = null then
			-- La clave buscada no est en el rbol
			return null;
		else
			if E = A.Clave then
				-- Se ha encontrado
				return A;
			else
				-- La bsqueda contina en el subrbol adecuado
				if E < A.Clave then
					return Buscar(A.Hijos(1),E);
				else
					return Buscar(A.Hijos(2),E);
				end if;
			end if;
		end if;
	end Buscar;

	procedure Vaciar(A: in out rbol_Rojo_Negro) is
	begin
		-- Para vaciar el rbol se recurre a un recorrido en postorden
		if A /= null then
			Vaciar(A.Hijos(1));
			Vaciar(A.Hijos(2));
			eliminar(A);
		end if;
	end Vaciar;

	function Tamao(A: in rbol_Rojo_Negro) return Natural is
	begin
		if A = null then
			-- Tamao del rbol vaco
			return 0;
		else
			-- Tamao del rbol no vaco
			return 1 + Tamao(A.Hijos(1)) + Tamao(A.Hijos(2));
		end if;
	end tamao;
	-------------------------------
	-- Operaciones constructoras --
	-------------------------------	
	procedure Insertar(C: in out Rojo_Negro; E: in TElemento) is
		Propiedad_Rojo: Boolean := False;
	begin
		Insertar(C.Elementos, C.Elementos, E, Propiedad_Rojo);
	end;

	procedure Extraer(C: in out Rojo_Negro;E: in TElemento) is
		Propiedad_Camino: Boolean := False;
	begin
		Extraer(C.Elementos, C.Elementos, E, Propiedad_Camino);
	end;

	procedure Vaciar(C: in out Rojo_Negro) is
	begin
		Vaciar(C.Elementos);
	end;

	------------------------------
	-- Operaciones observadoras --
	------------------------------
	function Est(C: in Rojo_Negro; E: in TElemento) return boolean is
	begin
		Return Buscar(C.Elementos, E) /= null;
	end;

	function Tamao(C: in Rojo_Negro) return Natural is
	begin
		return Tamao(C.Elementos);
	end;
	-----------------------------------------
	-- PRUEBA; BORRAR -----------------------
	-----------------------------------------
	procedure Propiedad_Rojo(A: in Rbol_Rojo_Negro) is
		Hijo, Otro: Rbol_Rojo_Negro;
	begin
		if A /= null then
			Propiedad_Rojo(A.Hijos(1));
			Hijo := A.Hijos(1);
			Otro := A.Hijos(2);
			if (A.Es_Rojo and ((Hijo /= null and then Hijo.Es_Rojo)
		 		or (Otro /= null and then Otro.Es_Rojo))) then
				raise Constraint_Error;
			end if;
			Propiedad_Rojo(A.Hijos(2));
		end if;
	end Propiedad_Rojo;

	function Propiedad_Camino(A: in Rbol_Rojo_negro) return Integer is
		hi, hd: Integer;
	begin
		if A = null then
			return 0;
		else
			hi := Propiedad_Camino(A.Hijos(1));
			hd := Propiedad_Camino(A.Hijos(2));
			if hi /= Hd then
				raise Constraint_Error;
			end if;

			if A.Es_Rojo then
				return Hi;
			else
				return Hi + 1;
			end if;
		end if;
	end Propiedad_Camino;

	procedure Comprobar_Rojo_Negro(C: in Rojo_Negro) is
		H: Natural;
	begin
		Propiedad_Rojo(C.Elementos);
	end Comprobar_Rojo_Negro;
	-----------------------------------------
	-- PRUEBA; BORRAR -----------------------
	-----------------------------------------
end Contenedores.Rojo_Negros;
