#ifndef INC_ROJONEGRO
#define INC_ROJONEGRO
#include "autoajustable.h"
#include "nodorojonegro.h"

template <class T> 
class RojoNegro:public AutoAjustable<T,NodoRojoNegro<T> > {
      NodoRojoNegro<T> *raiz;
      void rotarSimple(NodoRojoNegro<T> *&,int);
      void rotarDoble(NodoRojoNegro<T> *&,int);
      void recoloracion(NodoRojoNegro<T> *);
      bool hijosNegros(NodoRojoNegro<T> *);
      void insertarInterno(NodoRojoNegro<T>*&,const T&,bool &);
      void extraerInterno(NodoRojoNegro<T>*&,const T&,bool &);
      void extraerSucesor(NodoRojoNegro<T>*,NodoRojoNegro<T>*&,bool &);
      void restaurarPropiedadCamino(NodoRojoNegro<T> *&,int);
   public:
      RojoNegro(){raiz=NULL;}
      ~RojoNegro();
      void insertar(const T&);
      void extraer(const T&);
      bool encuentra(const T&)const;
      void vaciar();
};

// Funciones miembro de la clase RojoNegro
template <class T>
   RojoNegro<T>::~RojoNegro()
   {
      eliminarInterno(raiz);
   }

template <class T>
   void RojoNegro<T>::rotarSimple(NodoRojoNegro<T> *&nodo,int rama)
   {
      int otraRama=1-rama;
      bool auxColor;
      AutoAjustable<T, NodoRojoNegro<T> >::rotarSimple(nodo,rama);
      NodoRojoNegro<T> *hijo=nodo->enlaces[otraRama];
      //Se cambian los colores
      auxColor=nodo->esRojo;
      nodo->esRojo=hijo->esRojo;
      hijo->esRojo=auxColor;
   }

template <class T>
   void RojoNegro<T>::rotarDoble(NodoRojoNegro<T> *&nodo,int rama)
   {
      int otraRama=1-rama;
      AutoAjustable<T, NodoRojoNegro<T> >::rotarDoble(nodo,rama);
      NodoRojoNegro<T> *hijo=nodo->enlaces[otraRama];
      //Se cambian los colores
      nodo->esRojo=false;
      hijo->esRojo=true; 
   }

//Pone el color de un nodo a rojo y el de sus hijos a negro
template <class T>
   void RojoNegro<T>::recoloracion(NodoRojoNegro<T> *nodo)
   {
      NodoRojoNegro<T> *hijo=nodo->enlaces[0], *otroHijo=nodo->enlaces[1];
      nodo->esRojo=true;
      hijo->esRojo=false;
      otroHijo->esRojo=false;
   }

//Indica si los dos hijos de nodo son negros
template <class T>
   bool RojoNegro<T>::hijosNegros(NodoRojoNegro<T> *nodo)
   {
      NodoRojoNegro<T> *hijo=nodo->enlaces[0], *otroHijo=nodo->enlaces[1];
      if ((hijo==NULL)&&(otroHijo==NULL))
         return true;
      else
         if (((hijo!=NULL)&&(!(hijo->esRojo)))&&((otroHijo!=NULL)&&(!(otroHijo->esRojo))))
            return true;
         else
            return false;    
   }

template <class T>
   void RojoNegro<T>::insertarInterno(NodoRojoNegro<T>*&nodo,const T& valor,bool &propiedadRojo)
   {
      NodoRojoNegro<T> *padre, *tio, *hijo;
      int rama;
      if(nodo==NULL)
      {
        	//El nuevo valor se inserta como hoja
         nodo=new NodoRojoNegro<T>(valor);
         this->numElem++;
         propiedadRojo=true;
      }
      else
         //Si valor es igual a nodo->clave no se inserta
         if(valor!=nodo->clave)
         {
         	//La insercin continua en el subrbol adecuado
            if(valor<nodo->clave)
               rama=0;
            else
               rama=1;
            insertarInterno(nodo->enlaces[rama],valor,propiedadRojo);
            if(propiedadRojo)
               if(nodo->esRojo)   
               {
                  /*Hay que ascender otro nivel, a menos
                    que se est en la raz*/ 
                  if(nodo==raiz)  
                  {
                     //Slo se cambia el color si es la raz
                     nodo->esRojo=false;
                     propiedadRojo=false;
                  }
               }
               else
               {
                  //nodo es de color negro
                  padre=nodo->enlaces[rama];
                  //Se ha resuelto el conflicto rojo-rojo?  
                  if(hijosNegros(padre))      
                     propiedadRojo=false;
                  else
                  {
                     tio=nodo->enlaces[1-rama];
                     /*El color del to determina si se elige
                     una recoloracin o una rotacin */             
                     if((tio!=NULL) && (tio->esRojo)) 
                        recoloracion(nodo);
                     else
                     {
                        /*La relacin abuelo, padre e hijo es
                        izquierda-izquierda o derecha-derecha*/
                        hijo=padre->enlaces[rama];
                     	//Est ah el conflicto rojo-rojo?
                        if(hijo!=NULL && hijo->esRojo && padre->esRojo)
                           rotarSimple(nodo,rama);
                        else
                           rotarDoble(nodo,rama);
                        //Se han restablecido las propiedades del rbol Rojo-Negro  
                        propiedadRojo=false;
                     }
                  }  
               }  
         }    
   }

template <class T>
   void RojoNegro<T>::extraerInterno(NodoRojoNegro<T>*&nodo,const T& valor,bool &propiedadCamino)
   {
      int rama;
      NodoRojoNegro<T> *hijo, *hermano;
      if(nodo!=NULL)
      {
         if(valor==nodo->clave)
         /*Se ha encontrado el nodo a extraer
         tiene grado 0 o 1? */
            if((nodo->enlaces[0]==NULL)||(nodo->enlaces[1]==NULL))
            {
               if(nodo->enlaces[0]==NULL)
                  rama=1;
               else
                  rama=0;
               hijo=nodo->enlaces[rama];
               propiedadCamino=true;
               if(nodo->esRojo || nodo==raiz)
               /*El nodo que se va a borrar es rojo o
               es raz (con un solo hijo) */
                  propiedadCamino=false;
               else
                  if(hijo!=NULL && hijo->esRojo)
                  {
                     //El nodo s- es rojo
                     hijo->esRojo=false;
                     propiedadCamino=false;
                  }
               delete nodo;
               nodo=hijo;
               this->numElem--;
               return;
            }
            else
            {
            	/*Tiene grado 2, hay que extraer su sucesor
            	en simtrico */
               extraerSucesor(nodo,nodo->enlaces[1],propiedadCamino);
               rama=1;
            }
         else
         {
         	/*Si valor es distinto de nodo->clave, la extraccin se producir
         	en el subrbol que corresponda */
            if(valor<nodo->clave)
               rama=0;
            else
               rama=1;
            extraerInterno(nodo->enlaces[rama],valor,propiedadCamino);
         }
         //No se cumple la propiedad del camino
         if(propiedadCamino)
         { 
            hermano=nodo->enlaces[1-rama];
            if(!(hermano->esRojo) && hijosNegros(hermano))
            {
            	/*Caso 2 del texto (Se produce tambin dentro de
            	  Extraer_Sucesor) */
               hermano->esRojo=true;
               if(nodo->esRojo)
               {
                  //No se propaga la recoloracin
                  nodo->esRojo=false;
                  propiedadCamino=false;
               }
            }
            else
            {
               restaurarPropiedadCamino(nodo,rama);
               propiedadCamino=false;
            }
         }
      }
   }

template <class T>
   void RojoNegro<T>::extraerSucesor(NodoRojoNegro<T>*nodoExtraer,NodoRojoNegro<T>*&nodo,bool &propiedadCamino)
   {
      NodoRojoNegro<T> *hijo, *hermano;
      if(nodo->enlaces[0]==NULL)
      {
      	//nodo es el sucesor
         hijo=nodo->enlaces[1];
         nodoExtraer->clave=nodo->clave;
         propiedadCamino=true;
         if(nodo->esRojo)
            //El nodo que se va a borrar es rojo
            propiedadCamino=false;
         else
            if(hijo!=NULL && hijo->esRojo)
            {
            	//El nodo s- es rojo
               hijo->esRojo=false;
               propiedadCamino=false;
            }
         delete nodo;
         nodo=hijo;
         this->numElem--;
      }
      else
      {
      	//El sucesor es descendiente de nodo
         extraerSucesor(nodoExtraer,nodo->enlaces[0],propiedadCamino);
         if(propiedadCamino)
         { 
            hermano=nodo->enlaces[1];
            if(!(hermano->esRojo) && hijosNegros(hermano))
            {
            	//Caso 2 del texto
            	//(Se produce tambin en el cuerpo de extraer)
               hermano->esRojo=true;
               if(nodo->esRojo)
               {
                  //No se propaga la recoloracin 
                  nodo->esRojo=false;
                  propiedadCamino=false;
               }
            }
            else
            {
               restaurarPropiedadCamino(nodo,0);
               propiedadCamino=false;
            }
         }
      }
   }

template <class T>
   void RojoNegro<T>::restaurarPropiedadCamino(NodoRojoNegro<T> *&nodo,int rama)
   {
      NodoRojoNegro<T> *hermano, *hijoHerAlejadoS, *hijoHerProximoS;
      hermano=nodo->enlaces[1-rama];
      if(hermano->esRojo)
      {
      	//Caso 1 del texto
         rotarSimple(nodo,1-rama);
         hermano=nodo->enlaces[rama]->enlaces[1-rama];
         if(!(hermano->esRojo) && hijosNegros(hermano))
         {
            //Caso 2 del texto
            hermano->esRojo=true;
            nodo->enlaces[rama]->esRojo=false;
         	//Se han restablecido las propiedades del rbol Rojo-Negro
         }
         else
            restaurarPropiedadCamino(nodo->enlaces[rama],rama);
      }
      else
      {
         hijoHerAlejadoS=hermano->enlaces[1-rama];
         hijoHerProximoS=hermano->enlaces[rama];
         if(hijoHerAlejadoS!=NULL && hijoHerAlejadoS->esRojo)
         {
            //Caso 3 del texto
            rotarSimple(nodo,1-rama);
            hijoHerAlejadoS->esRojo=false;
         	//Se han restablecido las propiedades del rbol Rojo-Negro
         }
         else
            if((hijoHerAlejadoS==NULL || !(hijoHerAlejadoS->esRojo))&& hijoHerProximoS->esRojo)
            {
               //Caso 4 del texto
               rotarSimple(nodo->enlaces[1-rama],rama);
               restaurarPropiedadCamino(nodo,rama);
            }
      }
   }

template <class T>
   bool RojoNegro<T>::encuentra(const T& valor)const
   {
      return buscarInterno(raiz,valor);
   }

template <class T>
   void RojoNegro<T>::insertar(const T& valor)
   {
      bool propiedadRojo=false;
      insertarInterno(raiz,valor,propiedadRojo);
   }

template <class T>
   void RojoNegro<T>::extraer(const T& valor)
   {
      bool propiedadCamino=false;
      extraerInterno(raiz,valor,propiedadCamino);
   }

template <class T>
   void RojoNegro<T>::vaciar()
   {
      eliminarInterno(raiz);
      raiz=NULL;
      this->numElem=0;
   }

#endif



